import pandas as pd
import numpy as np
from tqdm import tqdm
from typing import Any

def create_polygons(adata: Any) -> Any:
    """
    Create polygons for each unique cell ID based on segmentation info.
    
    Parameters:
    - adata (AnnData): An AnnData object containing segmentation data under the key 'seg' in its 'uns' slot.

    Returns:
    - AnnData: The input AnnData object, updated with polygons for each cell under the key 'poly' in its 'uns' slot.
    """

    tenx_seg = adata.uns['seg']
    grouped = tenx_seg.groupby('cell_id')

    all_cells = sorted(list(set(tenx_seg.cell_id)))
 
    new_lib = [{
        'coordinates': [grouped.get_group(cell).iloc[:, 1:].to_numpy().tolist()],
        'type': 'Polygon'
    } for cell in tqdm(all_cells)]
 
    new_poly = dict(zip(map(str, all_cells), new_lib))
    adata.uns['poly'] = new_poly
    
    return adata



import random
from tqdm import tqdm
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import pandas as pd
from typing import Any, List, Tuple, Union, Dict, Optional
from scipy.sparse import csc_matrix
from shapely.geometry import Polygon

def normalize_data(adata, scale_factor=100):
    
    data_matrix = adata.X.toarray()
    
    total_counts = np.sum(data_matrix, axis=1).reshape(-1, 1)
    norm_mtx = (data_matrix / total_counts) * scale_factor
    log_mtx = np.log1p(norm_mtx)

    adata.X = csc_matrix(log_mtx.astype(np.float32))
    
    return adata
    
def to_hex(rgb_tuple: Tuple[float, float, float]) -> str:
    """
    Convert an RGB tuple to its hex color.
    
    Parameters:
    - rgb_tuple (Tuple[float, float, float]): A tuple representing RGB values (range 0 to 1).
    
    Returns:
    - str: Hexadecimal string representing the color.
    """
    r = int(rgb_tuple[0] * 255)
    g = int(rgb_tuple[1] * 255)
    b = int(rgb_tuple[2] * 255)
    hex_code = '#{0:02x}{1:02x}{2:02x}'.format(r, g, b)
    return hex_code


def crop(adata: Any, 
         xlims: Optional[Tuple[float, float]], 
         ylims: Optional[Tuple[float, float]]) -> Tuple[List[int], pd.DataFrame]:
    """
    Crop adata based on x and y limits.

    Parameters:
    - adata: An anndata object.
    - xlims: Tuple containing the x-axis limits.
    - ylims: Tuple containing the y-axis limits.

    Returns:
    - Tuple[List[int], pd.DataFrame]: A tuple containing the subset indices and the subsetted cell coordinates DataFrame.
    """
    
    cell_coord = adata.obs
    x_new = cell_coord.x_centroid.to_list()
    y_new = cell_coord.y_centroid.to_list()

    expand = 1.05
    if xlims is None:
        minx = cell_coord.x_centroid.min()
        maxx = cell_coord.x_centroid.max()
        xlims = [(minx + maxx) / 2.0 - (maxx - minx) / 2.0 * expand,
                 (minx + maxx) / 2.0 + (maxx - minx) / 2.0 * expand]

    if ylims is None:
        miny = cell_coord.y_centroid.min()
        maxy = cell_coord.y_centroid.max()
        ylims = [(miny + maxy) / 2.0 - (maxy - miny) / 2.0 * expand,
                 (miny + maxy) / 2.0 + (maxy - miny) / 2.0 * expand]

    x_lim_idx = [idx for idx, value in enumerate(x_new) if value > xlims[0] and value < xlims[1]]
    y_lim_idx = [idx for idx, value in enumerate(y_new) if value > ylims[0] and value < ylims[1]]

    subset_idx = list(set(x_lim_idx) & set(y_lim_idx))
    cell_coord2 = cell_coord.iloc[subset_idx]

    return subset_idx, cell_coord2


def get_color_map(adata: Any, 
                  color_by: Optional[str], 
                  cmap: Optional[Union[str, Dict[str, str]]], 
                  seed: int, 
                  genes: Optional[Union[str, List[str]]],
                  subset_idx: List[int]) -> Union[str, Dict[str, str]]:
    """
    Get a color map for the given adata.

    Parameters:
    - adata: An anndata object.
    - color_by (Optional[str]): Category from adata.obs by which to color the output.
    - cmap (Optional[Union[str, Dict[str, str]]]): Color map or a dictionary to map values to colors.
    - seed (int): Seed for random color generation.
    - genes (bool): Whether the map is based on genes or not.
    - subset_idx (List[int]): Indices of the subset of data to be considered.

    Returns:
    - Union[str, Dict[str, str]]: A color map or a dictionary mapping values to colors.
    """
    
    def generate_random_color():
        return "#" + ''.join([random.choice('0123456789ABCDEF') for _ in range(6)])

    def generate_cell_type_map():
        cell_type = list(adata.obs[color_by].cat.categories)
        key = f'{color_by}_colors'
        if key in adata.uns.keys():
            cell_colors = adata.uns[key]
        else:
            random.seed(seed)
            cell_colors = [generate_random_color() for _ in range(len(cell_type))]
        return dict(zip(cell_type, cell_colors))

    if not genes:
        if color_by is None:
            map_dict = ['#6699cc'] * len(subset_idx)
        elif cmap is None:
            map_dict = generate_cell_type_map()
        else:
            map_dict = cmap
    else:
        if cmap is None:
            colors = ['#f2f2f2', '#ffdbdb', '#fc0303']
            map_dict = matplotlib.colors.LinearSegmentedColormap.from_list("", colors)
        else:
            map_dict = cmap

    return map_dict


def split_field(df: pd.DataFrame, 
                n_fields_x: int, 
                n_fields_y: int, 
                x_col: str = 'x_centroid', 
                y_col: str = 'y_centroid') -> Tuple[List[List[Tuple[float, float]]], List[Tuple[float, float]]]:
    """
    Splits a spatial region into a grid of FOV based on x and y centroids in obs.
    
    Parameters:
    - df: AnnData.obs containing the x and y centroids.
    - n_fields_x: Number of FOV horizontally.
    - n_fields_y: Number of FOV vertically.
    - x_col: Column name for x centroids in df. Default is 'x_centroid'.
    - y_col: Column name for y centroids in df. Default is 'y_centroid'.
    
    Returns:
    - rectangles: each rectangle is represented by a list of its vertices.
    - centroids: centroids of these rectangles.
    """
    
    # Expand the bounding box slightly for better visualization
    expand = 1.1
    minx = df[x_col].min()
    maxx = df[x_col].max()
    x_range = expand * (maxx - minx)
    
    miny = df[y_col].min()
    maxy = df[y_col].max()
    y_range = expand * (maxy - miny)
    
    # Calculate the dimensions of each rectangle
    x_step = x_range / n_fields_x
    y_step = y_range / n_fields_y
    
    # Calculate the starting coordinates for the grid
    x_start = (minx + maxx) / 2.0 - (maxx - minx) / 2.0 * expand
    y_start = (miny + maxy) / 2.0 - (maxy - miny) / 2.0 * expand
    
    # Loop to create the grid of rectangles and calculate their centroids
    rectangles = []
    centroids = []
    for i in range(n_fields_x):
        for j in range(n_fields_y):
            top_left = (x_start + x_step * i, y_start + y_step * j)
            bottom_right = (top_left[0] + x_step, top_left[1] + y_step)
            centroid = (top_left[0] + x_step / 2, top_left[1] + y_step / 2)
    
            rectangle = [top_left, (bottom_right[0], top_left[1]), bottom_right, (top_left[0], bottom_right[1])]
            rectangles.append(rectangle)
            centroids.append(centroid)
    
    return rectangles, centroids

def filter_polygon(adata, poly_key = "poly", threshold = 10, view_summary = True):
    '''
    Filters polygons based on a size threshold
    
    Parameters:
    - adata: AnnData object.
    - threshold: Minimum area threshold for polygons to be kept.
    - view_summary: If True, displays a histogram of polygon areas.
    - inplace: If True, modifies `adata` in place, otherwise returns modified data.
    '''

    poly_dict = adata.uns.get(poly_key, {})
    polygon_areas = np.array([Polygon(poly['coordinates'][0]).area for poly in poly_dict.values()])
    
    if view_summary:
        print("Polygon area distribution")
        plt.figure(figsize=(5, 3))
        plt.hist(polygon_areas, bins=100, align='left', color='skyblue')
        plt.xlim(-10, 250)
        plt.gca().spines['top'].set_visible(False)
        plt.gca().spines['right'].set_visible(False)
        plt.tight_layout()
        plt.show()
        
    print(f"Original polygon number: {len(adata.uns['poly'])}")
    filtered_poly_dict = {pid: pd for pid, pd in poly_dict.items() if Polygon(pd['coordinates'][0]).area >= threshold}
    print(f"Keep polygon number: {len(filtered_poly_dict)}")
    
    print("Saving result...")
    keep_poly_number = [int(i) for i in filtered_poly_dict.keys()]
    adata.uns["seg"] = adata.uns["seg"][adata.uns["seg"].polygon_number.isin(keep_poly_number)]
    adata.uns[poly_key] = filtered_poly_dict
    adata = adata[adata.obs.polygon_number.isin(keep_poly_number)]
    return adata

def smooth_polygon(adata, window_size=5):
    """
    Smoothens the coordinates of polygons using a moving average filter.

    Parameters:
    - adata: AnnData object containing the dataset.
    - window_size: The size of the moving window for smoothing.
    """
    def smooth_coordinates(polygon, window_size):
        n = len(polygon)
        polygon = np.array(polygon)
        smoothed = polygon.copy()
        for i in range(n):
            indices = [(i + j - window_size // 2) % n for j in range(window_size)]
            smoothed[i] = np.mean(polygon[indices], axis=0)
        return smoothed.tolist()

    seg = adata.uns["seg"]
    all_polygons = sorted(set(seg["polygon_number"]))

    new_polygons = []
    for polygon_number in tqdm(all_polygons, desc="Smoothing polygons"):
        group = seg[seg.polygon_number == polygon_number]
        raw_coords = group.iloc[:, 3:5].values.tolist()
        smoothed_coords = smooth_coordinates(raw_coords, window_size = window_size)
        new_polygons.append({
            'coordinates': [smoothed_coords],
            'type': 'Polygon'
        })

    adata.uns['poly'] = dict(zip(map(str, all_polygons), new_polygons))
    

    












    
    


import pandas as pd
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.font_manager as fm
from matplotlib.colors import LinearSegmentedColormap
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
import math
from mpl_toolkits.axes_grid1 import make_axes_locatable
import random
from descartes import PolygonPatch
import scanpy as sc
from ._utils import to_hex, crop, get_color_map, split_field
from typing import Any, List, Optional, Union, Dict, Tuple

def plot_polygon(adata: Any, 
                 plot_type: str = "cell", 
                 xlims: Optional[Tuple[float, float]] = None, 
                 ylims: Optional[Tuple[float, float]] = None, 
                 color_by: Optional[str] = None, 
                 genes: Optional[Union[str, List[str]]] = None, 
                 seed: int = 123, 
                 alpha: float = 0.8,
                 cmap: Optional[Union[str, Dict[str, str]]] = None, 
                 ptsize: float = 0.01, 
                 ticks: bool = False, 
                 dpi: int = 300,
                 width: int = 5, 
                 height: int = 5, 
                 edgecolor: str = "#808080", 
                 linewidth: float = 0.2,
                 legend_col = 1) -> None:
    """
    Plot polygons based on the provided adata.

    Parameters:
    - adata: A anndata object.
    - plot_type: The type of plot ('cell', 'gene', or 'transcript').
    - xlims: Tuple containing the x-axis limits.
    - ylims: Tuple containing the y-axis limits.
    - color_by: Category from adata.obs by which to color the output.
    - genes: Genes to be considered for 'gene' or 'transcript' plots.
    - seed: Seed for random color generation.
    - alpha: Opacity of polygons.
    - cmap: Color map or a dictionary to map values to colors.
    - ptsize: Size of points in the scatter plot.
    - ticks: Whether to display axis ticks.
    - dpi: Resolution.
    - width: Width of the figure.
    - height: Height of the figure.
    - edgecolor: Color of the polygon edge.
    - linewidth: Width of the polygon edge line.

    Returns:
    None: Displays the desired plot based on the given parameters.
    """
    
    try:
        adata.uns['poly']
    except AttributeError:
        print('Please create polygons first! To create polygons, please use the create_polygons function.')
    else:
        poly_dict = {idx + 1: adata.uns['poly'][key] for idx, key in enumerate(adata.uns['poly'].keys())}
        new_lib = [*poly_dict.values()]

        subset_idx, new_coord = crop(adata, xlims, ylims)

        print("Total number of polygons: ", len(new_coord))

        if plot_type == "cell":
            if color_by is None:
                cols = ['#6699cc'] * len(subset_idx)
            else:
                map_dict = get_color_map(adata, color_by, cmap, seed, genes, subset_idx)
                cols = list(new_coord[color_by].map(map_dict))

            fig, ax = plt.subplots(dpi=dpi, figsize=(width, height))
            if not ticks:
                plt.tick_params(left=False, right=False, labelleft=False, labelbottom=False, bottom=False)
            [ax.add_patch(PolygonPatch(new_lib[i], fc=cols[j], ec=edgecolor, alpha=alpha, zorder=0.5, linewidth=linewidth)) for j, i in
             enumerate(subset_idx)]
            ax.axis('scaled')
            ax.grid(False)
            ax.invert_yaxis()
            
            scalebar = AnchoredSizeBar(ax.transData,
                       25, ' ', 'lower right', 
                       pad=0.5,
                       sep=5,
                       color='black',
                       frameon=False,
                       size_vertical=2,
                       fontproperties=fm.FontProperties(size=12))


            ax.add_artist(scalebar)
                
            if color_by is not None:
                markers = [plt.Line2D([0, 0], [0, 0], color=color, marker='o', linestyle='') for color in
                           map_dict.values()]
                ax.legend(markers, map_dict.keys(),
                          numpoints=1, loc='center left', bbox_to_anchor=(1, 0.5),
                          frameon=False, ncol=legend_col)
            plt.show()

        elif plot_type == "gene":
            if isinstance(genes, str):
                genes = [genes]

            def create_gene_plot(ax, gene, counts, cmap, subset_idx, alpha, edgecolor, ticks):
                # c_max = np.quantile(counts, 0.99)
                c_max = max(counts)
                bar_colors = [cmap(c / c_max) for c in counts]
                bar_colors = np.clip(bar_colors, 0, 1)
                all_colors = [to_hex(i) for i in bar_colors]
                cols = [all_colors[i] for i in subset_idx]
                if not ticks:
                    ax.tick_params(left=False, right=False, labelleft=False, labelbottom=False, bottom=False)

                ax.set_title(gene)
                [ax.add_patch(
                    PolygonPatch(new_lib[i], fc=cols[j], ec=edgecolor, alpha=alpha, zorder=0.1, linewidth=linewidth)) for j, i
                    in enumerate(subset_idx)]
                ax.axis('scaled')
                ax.grid(False)
                ax.invert_yaxis()
                
                scalebar = AnchoredSizeBar(ax.transData,
                       25, ' ', 'lower right', 
                       pad=0.5,
                       sep=5,
                       color='black',
                       frameon=False,
                       size_vertical=2,
                       fontproperties=fm.FontProperties(size=12))


                ax.add_artist(scalebar)
                norm = matplotlib.colors.Normalize(vmin=min(counts), vmax=c_max)

                return norm

            n_rows = math.ceil(len(genes) / 3)
            if len(genes) < 4:
                n_cols = len(genes)
            else:
                n_cols = 3

            fig, axs = plt.subplots(nrows=n_rows, ncols=n_cols, figsize=(n_cols * width, n_rows * height), dpi=dpi)
            # plt.subplots_adjust(hspace=0.5)

            if len(genes) == 1:
                counts = sc.get.obs_df(adata, keys=genes[0]).to_list()
                norm = create_gene_plot(axs, genes[0], counts, cmap, subset_idx, alpha, edgecolor, ticks)
                divider = make_axes_locatable(axs)
                cax = divider.append_axes("right", size="5%", pad=0.05)
                fig.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap), ax = axs, cax=cax)
            else:
                axs = axs.flatten()
                for gene, ax in zip(genes, axs):
                    counts = sc.get.obs_df(adata, keys=gene).to_list()
                    norm = create_gene_plot(ax, gene, counts, cmap, subset_idx, alpha, edgecolor, ticks)
                    divider = make_axes_locatable(ax)
                    cax = divider.append_axes("right", size="5%", pad=0.05)
                    plt.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap), ax=ax, cax=cax)

            if len(genes) < n_rows * n_cols:
                for idx in range(len(genes), n_rows * n_cols):
                    fig.delaxes(axs[idx])

            plt.tight_layout()
            plt.show()

        elif plot_type == "transcript":

            mol_data = adata.uns['transcript']
            cell_id_keep = new_coord.cell_id.tolist()
            mol_data2 = mol_data[mol_data.cell_id.isin(cell_id_keep) & mol_data.feature_name.isin(genes)]
            if cmap is None:
                random.seed(seed)
                tx_col = ["#" + ''.join([random.choice('0123456789ABCDEF') for j in range(6)])
                          for i in range(len(genes))]
                map_dict = {genes[i]: tx_col[i] for i in range(len(genes))}
            else:
                map_dict = cmap

            all_tx = mol_data2['feature_name'].to_list()
            all_tx = pd.DataFrame(all_tx, columns=['colors'])
            all_tx = all_tx.replace({'colors': map_dict})

            fig, ax = plt.subplots(dpi=dpi, figsize=(width, height))
            if not ticks:
                plt.tick_params(left=False, right=False, labelleft=False, labelbottom=False, bottom=False)
            [ax.add_patch(PolygonPatch(new_lib[i], fc='#fcfcfa', ec=edgecolor, alpha=alpha, zorder=0.5, linewidth=linewidth)) for j, i in
             enumerate(subset_idx)]
            scalebar = AnchoredSizeBar(ax.transData,
                       25, ' ', 'lower right', 
                       pad=0.5,
                       sep=5,
                       color='black',
                       frameon=False,
                       size_vertical=2,
                       fontproperties=fm.FontProperties(size=12))


            ax.add_artist(scalebar)
            
            ax.scatter(mol_data2.x_location, mol_data2.y_location, c=all_tx.colors, s=ptsize)
            ax.axis('scaled')
            ax.grid(False)
            ax.invert_yaxis()
            markers = [plt.Line2D([0, 0], [0, 0], color=color, marker='o', linestyle='') for color in map_dict.values()]
            ax.legend(markers, map_dict.keys(),
                      numpoints=1, loc='center left', bbox_to_anchor=(1, 0.5),
                      frameon=False, ncol=legend_col)
            plt.show()
            

def plot_scatter(adata: sc.AnnData, 
                 xlims: Optional[tuple] = None, 
                 ylims: Optional[tuple] = None, 
                 color_by: Optional[str] = None, 
                 genes: Optional[Union[str, List[str]]] = None, 
                 seed: int = 123, 
                 alpha: float = 0.8, 
                 cmap: Optional[str] = None, 
                 ptsize: float = 0.1, 
                 ticks: bool = False, 
                 dpi: int = 300, 
                 width: int = 5, 
                 height: int = 5, 
                 legend_loc: str = "center left", 
                 legend_col: int = 2, 
                 bbox_to_anchor: tuple = (1.0, 0.5), 
                 save: Optional[str] = None) -> None:
    """
    Plot scatter plots based on AnnData object.

    Parameters:
    - adata: A anndata object.
    - xlims, ylims: The x and y limits for cropping.
    - color_by: The column to color by.
    - genes: Gene(s) to plot.
    - seed: Random seed.
    - alpha: Transparency level.
    - cmap: Colormap.
    - ptsize: Point size.
    - ticks: Whether to show ticks.
    - dpi: Dots per inch.
    - width, height: Width and height of plot.
    - legend_loc: Legend location.
    - legend_col: Number of columns in the legend.
    - bbox_to_anchor: Legend bbox_to_anchor.
    - save: Filename to save the plot to.

    Returns:
    None
    """
    
    def plot(ax: plt.Axes, 
             group: str, 
             new_coord: pd.DataFrame, 
             alpha: float = 0.8, 
             ptsize: float = 0.01, 
             ticks: bool = False) -> None:
        """
        Helper function to create the scatter plot.
        """
        if not ticks:
            plt.tick_params(left=False, right=False, labelleft=False, labelbottom=False, bottom=False)
            ax.axis("off")

        ax.scatter(np.array(new_coord.x_centroid), np.array(new_coord.y_centroid), marker='o', linewidth=0,
                   alpha=alpha, color=np.array(new_coord['color']), s=ptsize)

        ax.set_title(group)
        ax.axis('scaled')
        ax.grid(False)
        ax.invert_yaxis()

    group = list(adata.obs.ident.cat.categories)

    n_rows = math.ceil(len(group) / 6)
    if len(group) < 6:
        n_cols = len(group)
    else:
        n_cols = 6

    fig, axs = plt.subplots(nrows=n_rows, ncols=n_cols, figsize=(n_cols * width, n_rows * height), dpi=dpi,
                            layout='constrained')

    if not genes:
        if len(group) == 1:
            subset_idx, new_coord = crop(adata, xlims, ylims)
            map_dict = get_color_map(adata, color_by, cmap, seed, genes, subset_idx)
            new_coord['color'] = new_coord[color_by].map(map_dict)
            plot(axs, group[0], new_coord, alpha, ptsize, ticks)
        else:
            axs = axs.flatten()
            for g, ax in zip(group, axs):
                tmp = adata[adata.obs.ident == g]
                subset_idx, new_coord = crop(tmp, xlims, ylims)
                map_dict = get_color_map(adata, color_by, cmap, seed, genes, subset_idx)
                new_coord['color'] = new_coord[color_by].map(map_dict)
                plot(ax, g, new_coord, alpha, ptsize, ticks)
        
            for ax in axs[len(group):]:
                ax.axis("off")
                ax.set_visible(False)
            
        markers = [plt.Line2D([0, 0], [0, 0], color=color, marker='o', linestyle='') for color in map_dict.values()]
        fig.legend(markers, map_dict.keys(), numpoints=1, loc=legend_loc, bbox_to_anchor=bbox_to_anchor, frameon=False,
                   ncol=legend_col)
        if save:
            plt.savefig(save, transparent=True, bbox_inches='tight')
            
        plt.show()

    else:
        if isinstance(genes, str):
            genes = [genes]

        if len(group) == 1:
            counts = sc.get.obs_df(adata, keys=genes[0]).to_list()
            subset_idx, new_coord = crop(adata, xlims, ylims)

            cmap = get_color_map(adata, color_by, cmap, seed, genes, subset_idx)

            # c_max = max(counts)
            c_max = np.quantile(counts, 0.99)
            c_min = min(counts)
            bar_colors = [cmap(c / c_max) for c in counts]
            bar_colors = np.clip(bar_colors, 0, 1)
            all_colors = [to_hex(i) for i in bar_colors]

            new_coord['color'] = [all_colors[i] for i in subset_idx]
            plot(axs, group[0], new_coord, alpha, ptsize, ticks)
            norm = matplotlib.colors.Normalize(vmin=c_min, vmax=c_max)
            cbar_ax = fig.add_axes([0.96, 0.2, 0.02, 0.6])
            fig.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap), cax=cbar_ax)
        else:
            c_max_list = []
            c_min_list = []
            
            axs = axs.flatten()
            for g, ax in zip(group, axs):
                tmp = adata[adata.obs.ident == g]
                counts = sc.get.obs_df(tmp, keys=genes[0]).to_list()
                c_max_list.append(np.quantile(counts, 0.99))
                c_min_list.append(min(counts))

            c_max = max(c_max_list)
            c_min = min(c_min_list)

            for g, ax in zip(group, axs.flatten()):
                tmp = adata[adata.obs.ident == g]
                counts = sc.get.obs_df(tmp, keys=genes[0]).to_list()
                subset_idx, new_coord = crop(tmp, xlims, ylims)

                cmap = get_color_map(adata, color_by, cmap, seed, genes, subset_idx)

                bar_colors = [cmap(c / c_max) for c in counts]
                bar_colors = np.clip(bar_colors, 0, 1)
                all_colors = [to_hex(i) for i in bar_colors]

                new_coord['color'] = [all_colors[i] for i in subset_idx]
                plot(ax, g, new_coord, alpha, ptsize, ticks)
            
            for ax in axs[len(group):]:
                ax.axis("off")
                
            norm = matplotlib.colors.Normalize(vmin=c_min, vmax=c_max)
            cbar_ax = fig.add_axes([1.02, 0.2, 0.005, 0.6])
            cb = fig.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap), cax=cbar_ax, orientation='vertical')
            '''
            if n_rows == 1:
                fig.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap), ax=axs[:], shrink=0.6, location="right")
            else:
                fig.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap), ax=axs[:, :], shrink=0.6, location="right")
            '''
        if save:
            plt.savefig(save)
            
        plt.show()

        
def plot_fov(data: "AnnData", 
             n_fields_x: int, 
             n_fields_y: int, 
             x_col: str = "x_centroid", 
             y_col: str = "y_centroid",
             group_label: Optional[str] = None, 
             highlight_cell: Optional[str] = None, 
             highlight_color: str = '#FC4B42',
             fill: bool = True, 
             point_size: float = 0.8, 
             alpha: float = 0.3,
             font_size: Optional[float] = None, 
             resolution: int = 200, 
             plot_width: int = 5, 
             plot_height: int = 5) -> None:
    """
    Plot field of view (FoV) with optional cell highlighting and field partitioning.
    
    Parameters:
    - adata: An anndata object.
    - n_fields_x, n_fields_y: Number of fields in x and y dimensions to partition the plot.
    - x_col, y_col: Column names indicating x and y coordinates in the data.
    - group_label: Column name to use for grouping and highlighting cells.
    - highlight_cell: Cell group label to be highlighted.
    - highlight_color: Color for highlighted cells.
    - fill: Whether to fill the polygons representing fields.
    - point_size: Size of the plotted cell points.
    - alpha: Transparency level for the polygons and cell points.
    - font_size: Font size for the field numbers. Calculated based on n_fields_x and n_fields_y if not provided.
    - resolution: Resolution of the plot in dpi.
    - plot_width, plot_height: Figure size.
    
    Returns:
    None
    """
    
    coord = data.obs.copy()
    
    rectangles, centroids = split_field(coord, n_fields_x, n_fields_y, x_col=x_col, y_col=y_col)
    
    # Convert rectangles to DataFrame for easier manipulation
    rectangle_df = pd.concat([pd.DataFrame(rect, columns=['x', 'y']).assign(id=idx) for idx, rect in enumerate(rectangles)])
    
    # Create polygons for plotting
    polygons = [{'coordinates': [group[['x', 'y']].values.tolist()], 'type': 'Polygon'} for _, group in rectangle_df.groupby('id')]
    
    if fill:
        fill_color = "#D3D3D3"
    else:
        fill_color = 'None'
    
    if font_size is None:
        font_size = 120 / (n_fields_x + n_fields_y)
    
    fig, ax = plt.subplots(dpi=resolution, figsize=(plot_width, plot_height))
    plt.tick_params(left=False, right=False, labelleft=False, labelbottom=False, bottom=False)
    
    if isinstance(group_label, str) and highlight_cell:
        coord['highlight'] = coord[group_label].apply(lambda x: highlight_color if x == highlight_cell else '#BFBFBF')
        [ax.plot(group_data[x_col], group_data[y_col], 'o', markersize=point_size, markerfacecolor=color, markeredgecolor='none', alpha=alpha) for color, group_data in coord.groupby('highlight')]
    else:
        ax.plot(coord[x_col].tolist(), coord[y_col].tolist(), 'o', markersize=point_size, markerfacecolor='#bfbfbf', markeredgecolor='none', alpha=alpha)
    
    for polygon in polygons:
        ax.add_patch(PolygonPatch(polygon, fc=fill_color, ec='black', alpha=alpha, linewidth=0.5))
    
    for idx, (x, y) in enumerate(centroids):
        ax.text(x, y, str(idx), fontsize=font_size, ha='center', va='center', weight='bold')
    
    ax.axis('scaled')
    ax.grid(False)
    ax.invert_yaxis()
    ax.set_frame_on(False)
    plt.show()
    
    
def subset_fov(adata: "AnnData", 
               fov: List[int], 
               n_fields_x: int, 
               n_fields_y: int, 
               x_col: str = 'x_centroid', 
               y_col: str = 'y_centroid') -> Tuple[float, float, float, float]:
    """
    Extract the coordinates that define the boundary of the specified field of view (FoV) from an AnnData object.

    Parameters:
    - adata: An anndata object.
    - fov: List of field of view indices to be considered.
    - n_fields_x, n_fields_y: Number of fields in x and y dimensions used to partition the plot.
    - x_col, y_col: Column names indicating x and y coordinates in the data.

    Returns:
    (xmin, xmax, ymin, ymax): Boundary coordinates in the format.
    """

    df = adata.obs.copy()
    rectangles, _ = split_field(df, n_fields_x, n_fields_y, x_col=x_col, y_col=y_col)
    
    selected_rectangles = [rectangles[i] for i in fov]
    min_pt = np.min(np.min(selected_rectangles, axis=0), axis=0)
    max_pt = np.max(np.max(selected_rectangles, axis=0), axis=0)
    
    return min_pt[0], max_pt[0], min_pt[1], max_pt[1]


import os
import anndata
import scanpy as sc
import pandas as pd
import numpy as np
from typing import Any
from tqdm import tqdm
from scipy.sparse import csr_matrix

def load_xenium(xenium_path: str) -> Any:
    """
    Load Xenium data into an AnnData object.
    
    Parameters:
    - xenium_path (str): Path to the Xenium output directory.

    Returns:
    - AnnData.
    """
    
    xenium_adata = sc.read_10x_h5(os.path.join(xenium_path, 'cell_feature_matrix.h5'))

    cell_coord = pd.read_csv(os.path.join(xenium_path, 'cells.csv.gz'), compression='gzip')
    molecule_data = pd.read_csv(os.path.join(xenium_path, 'transcripts.csv.gz'), compression='gzip')
    cell_seg = pd.read_csv(os.path.join(xenium_path, 'cell_boundaries.csv.gz'), compression='gzip')
    xenium_umap = pd.read_csv(os.path.join(xenium_path, 'analysis/umap/gene_expression_2_components/projection.csv'))
    xenium_clusters = pd.read_csv(os.path.join(xenium_path, 'analysis/clustering/gene_expression_graphclust/clusters.csv'))

    # Replace cell IDs that are -1 with 0
    molecule_data.loc[molecule_data['cell_id'] == -1, 'cell_id'] = 0
    
    xenium_adata.obs = cell_coord.copy()

    # Filter cells based on UMAP results
    cell_id_keep = xenium_umap.Barcode.tolist()
    xenium_adata = xenium_adata[xenium_adata.obs.cell_id.isin(cell_id_keep)]
    cell_seg = cell_seg[cell_seg.cell_id.isin(cell_id_keep)]

    # Update spatial information
    xenium_adata.obsm["spatial"] = xenium_adata.obs[["x_centroid", "y_centroid"]].copy().to_numpy()

    # Update UMAP data
    xenium_umap.index = xenium_adata.obs_names
    xenium_adata.obsm['X_umap'] = xenium_umap.iloc[:, 1:3].copy().to_numpy()

    # Update cluster information
    clusters = xenium_clusters['Cluster'].to_list()
    xenium_adata.obs['leiden'] = list(map(str, clusters))

    gene_names = xenium_adata.var_names.astype(str).tolist()
    molecule_data = molecule_data[molecule_data.feature_name.isin(gene_names)]
    molecule_data = molecule_data[molecule_data.cell_id.isin(cell_id_keep)]

    xenium_adata.uns['seg'] = cell_seg
    xenium_adata.uns['transcript'] = molecule_data
    xenium_adata.uns['datatype'] = "xenium"

    return xenium_adata

def load_xenium_baysor(data_dir):
    """
    Load Baysor segmented Xenium data.

    Parameters:
    - data_dir (str): Path to the Baysor output directory.

    Returns:
    - two Anndata object, one saves cells, the other saves polygons.
    """

    print("Loading baysor segmentation output...")
    counts = pd.read_csv(os.path.join(data_dir, "segmentation_counts.tsv"), sep="\t")

    cells_meta = pd.read_csv(os.path.join(data_dir, "segmentation_cell_stats.csv"))
    cells_meta.index = cells_meta.cell

    molecule_data = pd.read_csv(os.path.join(data_dir, "segmentation.csv"))
    molecule_data = molecule_data[molecule_data["cell"] != 0]
    molecule_data["x_location"] = molecule_data["x"] * 0.2125
    molecule_data["y_location"] = molecule_data["y"] * 0.2125
    
    features = list(counts["gene"])
    counts = csr_matrix(counts.iloc[:, 1:].values, dtype=np.int32)
    adata = anndata.AnnData(counts.T, obs=cells_meta)
    adata.var_names = features
    adata.uns["datatype"] = "xenium"

    adata.obs["x_centroid"] = adata.obs["x"] * 0.2125 ## convert pixel to micron
    adata.obs["y_centroid"] = adata.obs["y"] * 0.2125
    adata.obsm["spatial"] = adata.obs[["x_centroid", "y_centroid"]].copy().to_numpy()
    
    print("Loading polygons...")
    polygon_df = pd.read_csv(os.path.join(data_dir, "polygon.csv"))
    polygon_meta = pd.read_csv(os.path.join(data_dir, "polygon_meta.csv"))
    polygon_count = pd.read_csv(os.path.join(data_dir, "polygon_count.csv"))
    polygon_df["x_centroid"] = polygon_df["x_location"] * 0.2125
    polygon_df["y_centroid"] = polygon_df["y_location"] * 0.2125

    poly_counts = csr_matrix(polygon_count.iloc[:, 1:].values, dtype=np.int32)
    polygon_centroid = polygon_df.groupby('polygon_number').mean().reset_index()
    polygon_meta = polygon_meta.merge(polygon_centroid, on="polygon_number", how="inner")
    polygon_meta.index = polygon_meta.polygon_number.astype(str)
    poly_adata = anndata.AnnData(poly_counts, obs=polygon_meta)
    poly_adata.var_names = list(polygon_count.columns[1:])

    poly_adata.uns["seg"] = polygon_df
    poly_adata.uns["transcript"] = molecule_data

    print("Creating polygons...")
    seg = poly_adata.uns['seg']
    grouped = seg.groupby("polygon_number")
    all_polygons = sorted(list(set(seg["polygon_number"])))
    new_lib = [{
        'coordinates': [grouped.get_group(cell).iloc[:, 3:5].to_numpy().tolist()],
        'type': 'Polygon'
    } for cell in tqdm(all_polygons)]

    new_poly = dict(zip(map(str, all_polygons), new_lib))
    poly_adata.uns['poly'] = new_poly

    return adata, poly_adata

def load_polygon(data_dir):

    print("Loading baysor segmentation output...")
    molecule_data = pd.read_csv(os.path.join(data_dir, "segmentation.csv"))
    molecule_data = molecule_data[molecule_data["cell"] != 0]
    molecule_data["x_location"] = molecule_data["x"] * 0.2125
    molecule_data["y_location"] = molecule_data["y"] * 0.2125
    print("Loading polygons...")
    polygon_df = pd.read_csv(os.path.join(data_dir, "polygon.csv"))
    polygon_meta = pd.read_csv(os.path.join(data_dir, "polygon_meta.csv"))
    polygon_count = pd.read_csv(os.path.join(data_dir, "polygon_count.csv"))
    polygon_df["x_centroid"] = polygon_df["x_location"] * 0.2125
    polygon_df["y_centroid"] = polygon_df["y_location"] * 0.2125

    poly_counts = csr_matrix(polygon_count.iloc[:, 1:].values, dtype=np.int32)
    polygon_centroid = polygon_df.groupby('polygon_number').mean().reset_index()
    polygon_meta = polygon_meta.merge(polygon_centroid, on="polygon_number", how="inner")
    polygon_meta.index = polygon_meta.polygon_number.astype(str)
    poly_adata = anndata.AnnData(poly_counts, obs=polygon_meta)
    poly_adata.var_names = list(polygon_count.columns[1:])

    poly_adata.uns["seg"] = polygon_df
    poly_adata.uns["transcript"] = molecule_data

    print("Creating polygons...")
    seg = poly_adata.uns['seg']
    grouped = seg.groupby("polygon_number")
    all_polygons = sorted(list(set(seg["polygon_number"])))
    new_lib = [{
        'coordinates': [grouped.get_group(cell).iloc[:, 3:5].to_numpy().tolist()],
        'type': 'Polygon'
    } for cell in tqdm(all_polygons)]

    new_poly = dict(zip(map(str, all_polygons), new_lib))
    poly_adata.uns['poly'] = new_poly

    return  poly_adata

    
